/* 1 tab = 8 spaces! */

#include <windows.h>
#include <commctrl.h>

#include "global.h"
#include "version.h"
#include "resource.h"
#include "main.h"
#include "nes.h"
#include "draw.h"
#include "file.h"
#include "settings.h"
#include "sound.h"
#include "input.h"
#include "crystal.h"
#include "cont.h"
#include "mapper.h"
#include "help.h"


/* msg callback */
static LRESULT CALLBACK WndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch(msg)
	{
		case WM_DROPFILES: nes_drop(wParam); return FALSE;
		
		case WM_PAINT: case WM_SIZE: case WM_SIZING: case WM_MOVE: case WM_MOVING:
			 draw_repaint(); return FALSE;
		
		case WM_ACTIVATEAPP: MAIN->foreground=wParam; draw_repaint(); return FALSE;
		case WM_ENTERMENULOOP: case WM_ENTERSIZEMOVE: MAIN->dialog=TRUE; draw_repaint(); return FALSE;
		case WM_EXITMENULOOP: case WM_EXITSIZEMOVE: MAIN->dialog=FALSE; draw_repaint(); return FALSE;
		case WM_SYSKEYDOWN:
			if (input_menukey(wParam)) return FALSE;
			break;
		
		case WM_CLOSE: nes_stop(FALSE); KillTimer(MAIN->window,0); DestroyWindow(hwnd); MAIN->window=NULL; return FALSE;
		case WM_DESTROY: PostQuitMessage(0); return FALSE;
		
		case WM_SYSCOMMAND:
			switch (wParam&0xfff0) {
				case SC_KEYMENU:
					if (input_menukey(-1)) return FALSE;
					break;
				
				default: break;
			}
			break;
		
		case WM_LBUTTONDBLCLK: input_doubleclick(); return FALSE;
		
		/* the menu, bon appetit */
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDM_OPENGAME: nes_open(); break;
				case IDM_RELOADGAME: nes_reload(); break;
				case IDM_CLOSEGAME: nes_stop(TRUE); sound_silence(FALSE); draw_repaint(); break;
				case IDM_EXITEMU: PostMessage(hwnd,WM_CLOSE,0,0); break;
				
				case IDM_RESETNES: nes_set_action(NES_ACTION_RESET,0); break;
				case IDM_PAUSEEMU: nes_set_paused(nes_get_paused()^1); main_menu_check(IDM_PAUSEEMU,nes_get_paused()&1); main_menu_enable(IDM_RESETNES,(nes_get_paused()&1)^1); break;
				case IDM_SOUND: nes_set_action(NES_ACTION_SOUND,0); break;
				
				case IDM_ZOOM1: nes_set_action(NES_ACTION_ZOOM,1); break;
				case IDM_ZOOM2: nes_set_action(NES_ACTION_ZOOM,2); break;
				case IDM_ZOOM3: nes_set_action(NES_ACTION_ZOOM,3); break;
				case IDM_ZOOM4: nes_set_action(NES_ACTION_ZOOM,4); break;
				case IDM_FULLSCREEN: nes_set_action(NES_ACTION_SCREENMODE,0); break;
				case IDM_CORRECTASPECT: nes_set_action(NES_ACTION_CORRECTASPECT,0); break;
				case IDM_STRETCHFIT: nes_set_action(NES_ACTION_STRETCHFIT,0); break;
				
				case IDM_RENDERD3D: if (draw_get_rendertype()!=DRAW_RENDERTYPE_D3D) { draw_set_rendertype(DRAW_RENDERTYPE_D3D); draw_set_recreate(); } break;
				case IDM_RENDERDD: if (draw_get_rendertype()!=DRAW_RENDERTYPE_DD) { draw_set_rendertype(DRAW_RENDERTYPE_DD); draw_set_recreate(); } break;
				case IDM_SOFTRENDER: draw_set_softrender(draw_get_softrender()^1); draw_set_recreate(); break;
				case IDM_BFILTER: draw_set_bfilter(draw_get_bfilter()^1); break;
				case IDM_FLIP_X: draw_set_flip_x(draw_get_flip_x()^1); break;
				case IDM_VSYNC: draw_set_vsync(draw_get_vsync()^1); if (draw_get_rendertype()==DRAW_RENDERTYPE_D3D) draw_set_recreate(); break;
				case IDM_VIDEOFORMAT: nes_set_action(NES_ACTION_VIDEOFORMAT,0); break;
				
				case IDM_VSMONOCHROME: draw_set_surface_change(DRAW_SC_VIDSIGNAL|DRAW_VIDSIGNAL_MONOCHROME); break;
				case IDM_VSCOMPOSITE: draw_set_surface_change(DRAW_SC_VIDSIGNAL|DRAW_VIDSIGNAL_COMPOSITE); break;
				case IDM_VSSVIDEO: draw_set_surface_change(DRAW_SC_VIDSIGNAL|DRAW_VIDSIGNAL_SVIDEO); break;
				case IDM_VSRGB: draw_set_surface_change(DRAW_SC_VIDSIGNAL|DRAW_VIDSIGNAL_RGB); break;
				case IDM_VSCUSTOM: MAIN->dialog=TRUE; DialogBox(MAIN->module,MAKEINTRESOURCE(IDD_VIDEOSIGNAL),MAIN->window,draw_change_vidsignal_custom); MAIN->dialog=FALSE; break;
				
				case IDM_HELP: MAIN->dialog=TRUE; DialogBox(MAIN->module,MAKEINTRESOURCE(IDD_HELP),MAIN->window,help_help); MAIN->dialog=FALSE; break;
				case IDM_ABOUT: MAIN->dialog=TRUE; DialogBox(MAIN->module,MAKEINTRESOURCE(IDD_ABOUT),MAIN->window,help_about); MAIN->dialog=FALSE; break;
				
				
				case IDM_CPUSPEEDPLUS: nes_set_action(NES_ACTION_CPUSPEEDPLUS,0); break;
				case IDM_CPUSPEEDMIN: nes_set_action(NES_ACTION_CPUSPEEDMIN,0); break;
				case IDM_CPUSPEEDNORMAL: nes_set_action(NES_ACTION_CPUSPEEDNORMAL,0); break;
				
				case IDM_CARTSWITCHUP: nes_set_action(NES_ACTION_CARTSWITCHUP,0); break;
				case IDM_CARTSWITCHDOWN: nes_set_action(NES_ACTION_CARTSWITCHDOWN,0); break;
				
				default: break;
			}
			return FALSE;
		
		/* timer timeout */
		case WM_TIMER:
			WaitForSingleObject(MAIN->mutex,INFINITE);
			
			/* workaround: WM_ACTIVATEAPP should be enough to determine window priority, but for some reason it gets rarely ignored */
			MAIN->foreground=MAIN->window==GetForegroundWindow();
			
			input_mouse_cursor_update();
			draw_check_reset(); /* resetting/recreating the renderer device is the responsibility of the thread that created it */
			
			if (!nes_is_running()||nes_get_paused()) { draw_draw(); input_read(); input_read_shortcuts(); }
			ReleaseMutex(MAIN->mutex);
			return FALSE;
		
		default: break;
	}
	
	return DefWindowProc(hwnd,msg,wParam,lParam);
}

/* program entry */
int WINAPI WinMain(HINSTANCE inst,HINSTANCE prev,LPSTR arg,int show)
{
	WNDCLASSEX wc;
	MSG msg;
	RECT rect;
	HGDIOBJ bgc_start=GetStockObject(WHITE_BRUSH);
	
	memset(&wc,0,sizeof(WNDCLASSEX));
	
	InitCommonControls();
	
	log_init();
	crystal_init();
	draw_setzero();
	draw_create_text_mutex();
	draw_set_text_max_ticks_default();
	
	if ((CoInitializeEx(NULL,COINIT_APARTMENTTHREADED))!=S_OK) {
		LOG(LOG_MISC|LOG_ERROR,"can't initialize COM!\n");
		exit(1);
	}
	
	timeBeginPeriod(1);
	
	MEM_CREATE(MAIN,sizeof(Main));
	MAIN->module=inst;
	
	#if DEBUG_PROFILER
	PROFILER_INIT();
	#endif
	
	file_init();
	settings_init();
	file_directories();
	
	if ((MAIN->mutex=(CreateMutex(NULL,FALSE,NULL)))==NULL) { LOG(LOG_MISC|LOG_ERROR,"can't create main mutex!\n"); exit(1); }
	
	/* register the window class */
	wc.cbSize=sizeof(WNDCLASSEX);			/* struct size */
	wc.lpfnWndProc=WndProc;				/* pointer to wndproc */
	wc.hInstance=MAIN->module;			/* handle to app instance */
	wc.hIconSm=wc.hIcon=(HICON)LoadImage(MAIN->module,MAKEINTRESOURCE(ID_ICON),IMAGE_ICON,0,0,LR_DEFAULTSIZE); /* 32x32 icon (alt-tab), and small icon (top left, file) */
	wc.hCursor=(HCURSOR)LoadImage(NULL,MAKEINTRESOURCE(OCR_NORMAL),IMAGE_CURSOR,0,0,LR_SHARED); /* cursor when it's over the window */
	wc.hbrBackground=(HBRUSH)bgc_start;		/* bg colour */
	wc.lpszMenuName=MAKEINTRESOURCE(ID_MENU);	/* name of menu resource */
	wc.lpszClassName=version_get(VERSION_NAME);	/* class name */
	wc.style=CS_DBLCLKS;				/* style */
	
	if(!RegisterClassEx(&wc)) { LOG(LOG_MISC|LOG_ERROR,"can't register window!\n"); exit(1); }
	
	
	#define WIN_STYLE	WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN
	#define WIN_EX_STYLE	WS_EX_CLIENTEDGE|WS_EX_ACCEPTFILES
	
	/* get (window size)-(client area size) */
	rect.left=rect.top=rect.right=rect.bottom=0;
	AdjustWindowRectEx(
		&rect,		/* location of rect struct */
		WIN_STYLE,	/* window style */
		TRUE,		/* has menu? */
		WIN_EX_STYLE	/* extended style */
	);
	MAIN->x_plus=rect.right-rect.left;
	MAIN->y_plus=rect.bottom-rect.top;
	
	MAIN->menu=LoadMenu(MAIN->module,MAKEINTRESOURCE(ID_MENU));
	
	/* create window */
	MAIN->window=CreateWindowEx (
			WIN_EX_STYLE,			/* extended style */
			wc.lpszClassName,		/* class name */
			version_get(VERSION_NAME),	/* window title */
			WIN_STYLE,			/* window style */
			CW_USEDEFAULT,			/* x placement */
			CW_USEDEFAULT,			/* y placement */
			MAIN->x_plus+256*2,		/* width */
			MAIN->y_plus+224*2,		/* height */
			NULL,				/* parent window handle */
			MAIN->menu,			/* menu handle */
			MAIN->module,			/* app instance handle */
			NULL				/* pointer to window creation data */
	);
	
	if(MAIN->window==NULL) { LOG(LOG_MISC|LOG_ERROR,"can't create window!\n"); exit(1); }
	draw_init(); main_center_window(MAIN->window);
	crystal_settings_load();
	switch (draw_get_vidformat()) {
		case DRAW_VIDFORMAT_AUTO: case DRAW_VIDFORMAT_NTSC: crystal_set_mode(0); break;
		case DRAW_VIDFORMAT_PAL: crystal_set_mode(1); break;
		default: break;
	}
	
	ShowWindow(MAIN->window,show);
	UpdateWindow(MAIN->window);
	
	LOG(LOG_MISC|LOG_COLOUR(1),"Welcome to %s %s,  %s\n",version_get(VERSION_NAME),version_get(VERSION_NUMBER),version_get(VERSION_AUTHOR));
	
	draw_draw();
	SetClassLongPtr(MAIN->window,GCL_HBRBACKGROUND,0);
	DeleteObject(bgc_start);
	
	sound_init();
	nes_init();
	input_init();
	cont_init();
	
	SetTimer(MAIN->window,0,15,NULL);
	
	if (strlen(arg)) {
		file_setfile(NULL,arg,NULL,"nes");
		nes_start();
	}
	
	
	/* messageloop */
	while(GetMessage(&msg,NULL,0,0)>0) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	
	
	settings_save();
	
	nes_clean();
	draw_clean();
	
	if (sound_get_sleepthread()!=NULL) {
		WaitForSingleObject(sound_get_sleepthread(),INFINITE);
		CloseHandle(sound_get_sleepthread());
	}
	
	sound_clean();
	cont_clean();
	input_clean();
	mapper_clean_info();
	settings_clean();
	file_clean();
	
	timeEndPeriod(1);
	
	CoUninitialize();
	
	DestroyMenu(MAIN->menu);
	DestroyIcon(wc.hIconSm);
	DestroyIcon(wc.hIcon);
	CloseHandle(MAIN->mutex);
	
	UnregisterClass(wc.lpszClassName,MAIN->module);
	
	MEM_CLEAN(MAIN);
	/* LOG_ERROR unallowed from here */
	LOG(LOG_VERBOSE,"main cleaned\n");
	crystal_clean();
	
	return msg.wParam;
}

void main_menu_enable(UINT id,int en) { EnableMenuItem(MAIN->menu,id,en==FALSE?MF_GRAYED:MF_ENABLED); DrawMenuBar(MAIN->window); }
void main_menu_check(UINT id,int en) { CheckMenuItem(MAIN->menu,id,en==FALSE?MF_UNCHECKED:MF_CHECKED); DrawMenuBar(MAIN->window); }
void main_menu_radio(UINT id,UINT first,UINT last) { CheckMenuRadioItem(MAIN->menu,first,last,id,MF_BYCOMMAND); DrawMenuBar(MAIN->window); }

void main_menu_caption(int get,UINT id,char* c)
{
	MENUITEMINFO m;
	memset(&m,0,sizeof(MENUITEMINFO));
	m.cbSize=sizeof(MENUITEMINFO);
	m.fMask=MIIM_STRING;
	m.dwTypeData=c;
	if (get) { m.cch=STRING_SIZE; GetMenuItemInfo(MAIN->menu,id,FALSE,&m); }
	else { SetMenuItemInfo(MAIN->menu,id,FALSE,&m); DrawMenuBar(MAIN->window); } /* drawmenubar doesn't seem to have effect */
}


/* center window on its parent */
void main_center_window(HWND window)
{
	RECT rect_parent,rect_window,rect;
	int x,y;
	
	HWND parent=GetParent(window);
	
	memset(&rect_parent,0,sizeof(RECT));
	memset(&rect_window,0,sizeof(RECT));
	memset(&rect,0,sizeof(RECT));
	
	if (parent==NULL) { parent=GetDesktopWindow(); GetClientRect(parent,&rect_parent); }
	else GetWindowRect(parent,&rect_parent);
	GetWindowRect(window,&rect_window);
	CopyRect(&rect,&rect_parent);
	
	OffsetRect(&rect_window,-rect_window.left,-rect_window.top);
	OffsetRect(&rect,-rect.left,-rect.top);
	OffsetRect(&rect,-rect_window.right,-rect_window.bottom);
	
	x=rect_parent.left+(rect.right/2.0);
	y=rect_parent.top+(rect.bottom/2.0);
	if (x>(GetSystemMetrics(SM_CXFULLSCREEN)-(rect_window.right-rect_window.left))) x=GetSystemMetrics(SM_CXFULLSCREEN)-(rect_window.right-rect_window.left);
	if (y>(GetSystemMetrics(SM_CYFULLSCREEN)-(rect_window.bottom-rect_window.top))) y=GetSystemMetrics(SM_CYFULLSCREEN)-(rect_window.bottom-rect_window.top);
	if (x<0) x=0; if (y<0) y=0;
	
	SetWindowPos(window,NULL,x,y,0,0,SWP_NOSIZE|SWP_NOZORDER);
}

/* set titlebar text */
void main_titlebar(const char* text)
{
	char title[STRING_SIZE]={0};
	
	if (text==NULL) SetWindowText(MAIN->window,version_get(VERSION_NAME));
	else {
		sprintf(title,"%s - %s",version_get(VERSION_NAME),text);
		SetWindowText(MAIN->window,title);
	}
}
