/* Nintendo mappers */

static struct {
	/* nrom copynes */
	int nrom_copynes_reset;
	
	/* cnrom with chr disable */
	int cnrom_dis_vrom_disabled;
	
} map_nintendo;


/* --- NES-AOROM/NES-AMROM/NES-ANROM (iNES: 7) ---
	Solstice, Marble Madness, Wizards & Warriors, Battletoads, etc.
*/

_MAPIO(aorom)
{
	/* bits 0,1,2,3: swap 32k prgbank in $8000 */
	int bank=data<<2&MAXPRG;
	PRG32_BS(bank);
	
	/* bit 4: set mirroring */
	MIR_01(data&BIT(4))
}

_MAPINIT(aorom)
{
	SET_MAPIO8_ALL(aorom);
	INIT_PRG_FIRST();
	INIT_CHR_FIRST(); /* 8k vram */
	INIT_MIR_0();
}





/* --- NES-BNROM (iNES: 34) ---
	Deadly Towers/Mashou, Caesars Palace, Classic Concentration, Darkseed, several hacks
*/

_MAPIO(bnrom)
{
	/* select 32k prg bank */
	int prg=data<<2&MAXPRG;
	PRG32_BS(prg);
}

_MAPINIT(bnrom)
{
	SET_MAPIO8_ALL(bnrom);
	INIT_PRG_FIRST();
	INIT_CHR_FIRST();
}





/* --- NES-CNROM (iNES: 3) ---
	Solomon's Key, Gradius, Hudson's Adventure Island, Puzznic, Legend of Kage, etc.
	
   --- CNROM relying on bus conflicts (iNES: no/3) ---
	Cybernoid - The Fighting Machine
	
   --- CNROM with Bandai speech chip (iNES: no/3) ---
	Family Trainer - Aerobics Studio (Japanese)
	
   --- CNROM with CHR ROM disable (not sure if it's a Nintendo board) (iNES: 185) ---
	Sansuu 1 Nen,2,3, Banana, B-Wings, Bird Week, Mighty Bomb Jack, Seicross, Spy vs Spy
*/

/* TODO:
	Family Trainer - Fuuun! Takeshi Shiro 2 (J): doesn't work, probably a bad dump
	Banana (J) (chr disable) in GoodNES is probably a bad dump, it misses a 2nd VROM bank
	
	- Family Trainer - Aerobics Studio: correct sound emulation using an external ROM,
	- find out correct behaviour for CNROM with chr disable, though everything works fine this way
*/

/* standard */
_MAPIO(cnrom)
{
	/* swap 8k vrom bank in PPU $0000 */
	int bank=data<<3&MAXVROM;
	CHR8_BS_V(bank);
}

/* bus conflicts */
_MAPIO(cnrom_bus) { int bank=(data&bus)<<3&MAXVROM; CHR8_BS_V(bank); } /* almost same as normal */

/* chr disable */
_MAPIO(cnrom_dis)
{
	MAPIO_FUN(cnrom)(address,bus,data);
	
	/* disable/enable vrom, method guessed */
	map_nintendo.cnrom_dis_vrom_disabled^=1;
}
_MAPPPUBUS(cnrom_dis) { if (map_nintendo.cnrom_dis_vrom_disabled&!*mapper.ppu_rendering_ptr) *mapper.ppu_memmap_buffer_ptr=0x12; }

/* bandai speech chip ($6000) */
_MAPIO(cnrom_bs)
{
	/* bit 6: play/init sound? bits 0,1,2: sample number */
	if (!(data&BIT(6))) {
		#if DEBUG_MAPPER
		LOG(LOG_MAPPER,"%02x --- %2d\n",data,data&7);
		#endif
		sample_play(data&7);
	}
}

_MAPINIT(cnrom)
{
	INIT_PRG_FIRST();
	INIT_CHR_FIRST(); /* first 8k vrom */
	SET_MAPIO8_ALL(cnrom); /* std */
	
	switch (maptype) {
		/* std */
		case MAPTYPE_NINTENDO_CNROM:
			break;
		
		/* bus conflicts */
		case MAPTYPE_NINTENDO_CNROM_RELYING_ON_BUS_CONFLICTS:
			SET_MAPIO8_ALL(cnrom_bus);
			break;
		
		/* chr disable */
		case MAPTYPE_NINTENDO_CNROM_WITCH_CHR_DISABLE:
			SET_MAPIO8_ALL(cnrom_dis);
			
			/* set ppu patterntable read overrides */
			ppu_set_fpt(PPU_FPT_PT0,FALSE,ppu_read_pt0a_override,ppu_read_pt0b_override,ppu_read_pt0c_override,ppu_read_pt0d_override,NULL,NULL,NULL,NULL);
			ppu_set_fpt(PPU_FPT_PT1,FALSE,ppu_read_pt1a_override,ppu_read_pt1b_override,ppu_read_pt1c_override,ppu_read_pt1d_override,NULL,NULL,NULL,NULL);
			SET_MAPPPUBUS(cnrom_dis);
			break;
		
		/* bandai speech chip */
		case MAPTYPE_NINTENDO_CNROM_WITH_BANDAI_SPEECH_CHIP:
			SET_MAPIO6(cnrom_bs);
			sample_load();
			SET_MAPSTREAM(sample);
			break;
		
		default: break;
	}
}





/* --- NES-CPROM (iNES: 13) ---
	Videomation
*/

_MAPIO(cprom)
{
	/* select 4k vram for ppu $1000 */
	data=data<<2&0xf;
	CHR4_BS(4,data,vram);
}

_MAPINIT(cprom)
{
	int i;
	SET_MAPIO8_ALL(cprom);
	INIT_PRG_FIRST(); /* first (and only) 32k prgbank */
	INIT_CHR_FIRST(); /* it's got 16k vram, 1st 4k fixed */
	
	/* allocate another 8k vram bank, cleaned in cartridge_unload */
	for (i=8;i<0x10;i++) { MEM_CREATE(cartridge->vram[i],0x400); }
}





/* --- NES-GNROM/NES-MH (iNES: 66) ---
	AV - Mahjongg (actually FFE?), Dragon Ball - Le Secret Du Dragon (F)/Dragon Ball - Shen Long no Nazo (J)/Dragon Power (U),
	Family Block, Gumshoe, Jajamaru no Daibouken, Mississippi Satsujin Jiken, Mobile Suit Z Gundam - Hot Scramble,
	Paris-Dakar Rally Special, Super Mario Bros. + Duck Hunt, Takahashi Meijin no Bugutte Honey, Thunder & Lightning,
*/

_MAPIO(gnrom)
{
	/* low nibble: select 8k vrom bank */
	int bank=data<<3&BIN8(01111000)&MAXVROM;
	if (MAXVROM) { CHR8_BS_V(bank); }
	
	/* high nibble: select 32k prg bank */
	bank=data>>2&BIN8(00111100)&MAXPRG;
	PRG32_BS(bank);
}

_MAPINIT(gnrom)
{
	SET_MAPIO8_ALL(gnrom);
	INIT_PRG_FIRST();
	INIT_CHR_FIRST();
}





/* --- NES-NROM-128/NES-NROM-256/NES-RROM-128 (iNES: 0) ---
	Super Mario Bros, Ice Climber, Pac-Man, Donkey Kong, Dig Dug, Millipede, Pooyan, Galaga, etc.
	
   --- NROM Family Basic (back up switch on cartridge, sram with 2 AA batteries) (iNES: no/0) ---
	Family Basic, Playbox Basic *** both Japanese
	
   --- NROM with CopyNES (iNES: no/0) ---
	several PD programs that (overseen by author) expect the NES to have been initialised, eg:
	pputime2, Sayoonara, Sokoban, Stars Demo, Stars SE
*/

/* standard has no IO */

/* familybasic */
_MAPIO(nrom_fb_wram)
{
	/* if back up switch is on, wram below $7000 is write protected */
	if ((address<0x7000)&mapper_switches.familybasic_backupswitch) return;
	WRAM_WRITE(data);
}

_MAPSWITCHINIT(nrom_fb)
{
	int i=settings_get_yesnoauto(SETTINGS_FBSWITCH);
	if (i==FALSE||i==TRUE) mapper_switches.familybasic_backupswitch=i;
	else mapper_switches.familybasic_backupswitch=TRUE;
}

_MAPSWITCHSAVE(nrom_fb)
{
	SETTINGS_PUT_STRING(settings_info(SETTINGS_FBSWITCH),mapper_switches.familybasic_backupswitch?"on":"off");
}

_MAPSWITCHSET(nrom_fb)
{
	switch (set&7) {
		case MAPSWITCH_UP: set=1; break;
		case MAPSWITCH_DOWN: set=0; break;
		default: return;
	}
		
	if (set!=mapper_switches.familybasic_backupswitch) {
		LOG(LOG_MISC|LOG_COLOUR(3)|LOG_TYPE(4),"back up switch O%s",set?"N ":"FF");
		mapper_switches.familybasic_backupswitch=set;
	}
}

/* copynes */
_MAPRESET(nrom_copynes)
{
	/* first reset, CopyNES initialisation */
	if (!map_nintendo.nrom_copynes_reset) {
		map_nintendo.nrom_copynes_reset=TRUE;
		apu_write_fs(0x40); /* disable frame irq, not sure */
		cpu_memmap_fill_ram(0);
	}
}

/* init */
_MAPINIT(nrom)
{
	INIT_PRG_FIRST(); /* 32k?:full, 16k?:mirrored */
	INIT_CHR_FIRST();
	
	switch (maptype) {
		/* standard */
		case MAPTYPE_NINTENDO_NROM: break;
		
		/* family basic */
		case MAPTYPE_NINTENDO_NROM_FAMILY_BASIC:
			DEF_6000_WRAM();
			SET_MAPIO6(nrom_fb_wram);
			SET_MAPSWITCHSET(nrom_fb);
			break;
		
		/* copynes */
		case MAPTYPE_NINTENDO_NROM_WITH_COPYNES:
			SET_MAPRESET(nrom_copynes);
			break;
		
		default: break;
	}
}





/* --- NES-UNROM/NES-UOROM (iNES: 2) ---
	Castlevania, Ghosts 'n Goblins, Rockman, Rodland, Ducktales 1,2, Contra, Metal Gear, Guardian Legend, etc.
*/

_MAPIO(unrom)
{
	/* swap 16k prgbank in 0x8000 */
	int bank=data<<1&MAXPRG;
	PRG16_BS(0,bank);
}

_MAPINIT(unrom)
{
	SET_MAPIO8_ALL(unrom);
	INIT_PRG_FIRSTLAST();
	INIT_CHR_FIRST(); /* 8k vram (some games use VROM with this mapper, though that's wrong) */
}
